// Copyright (C) 2019  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Package δbtree provides BTree diffing.
package δbtree

//go:generate ../../gen-set δbtree Tree *Tree zset_tree.go
//go:generate ../../gen-set δbtree Object interface{} zset_object.go

import (
	"fmt"

	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
)

type Tree   = btree.LOBTree
type Bucket = btree.LOBucket

// PathSet is a collection of BTree paths.
//
// For example it can be a set of visited BTree paths.
// There is no requirement that paths belong to only one single BTree.
//
// PathSet is not safe for concurrent access.
type PathSet struct {
	// residency index: BTree|Bucket to top tree element.
	residencyIdx map[zodb.Oid]SetTree // oid -> {} roots

	// master index: tree root -> {} masters
	masterIdx map[*Tree]SetObject
}

// NewPathSet creates new empty PathSet.
func NewPathSet() *PathSet {
	return &PathSet{
		residencyIdx: make(map[zodb.Oid]SetTree),
		masterIdx:    make(map[*Tree]SetObject),
	}
}

// Add adds path to collection and associates path root with master.
//
// path[0] signifies a root.
// All path elemens must be Tree except last one which must be Bucket.
// A root can be associated with several masters (each provided on different Add call)
//
// XXX catch cycles on add?
func (m *PathSet) Add(master interface{}, path []zodb.IPersistent) {
	//fmt.Printf("\n\nPathSet.Add %v %s\n", master, path)

	l := len(path)
	if l == 0 {
		panic("empty path")
	}

	var root *Tree
	for i, node := range path {
		switch node := node.(type) {
		default:
			panic(fmt.Sprintf("path elements must be Tree|Bucket, not %T", node))

		case *Tree:
			if i == 0 {
				root = node
			}
			// XXX check only not last

		case *Bucket:
			// XXX check only last

		}
	}

	for _, node := range path {
		nodeRoots, ok := m.residencyIdx[node.POid()]
		if !ok {
			nodeRoots = make(SetTree)
			m.residencyIdx[node.POid()] = nodeRoots
		}
		nodeRoots.Add(root)
	}

	masterSet, ok := m.masterIdx[root]
	if !ok {
		masterSet = SetObject{}
		m.masterIdx[root] = masterSet
	}
	masterSet.Add(master)
}

// Invalidates returns which masters are invalidated by changed objects.
//
// returned are: roots of changed trees.
//
// XXX stub.
//
// TODO toposort changev wrt nodes ordering (defined by traversed paths) and do δbtree top-down.
func (m *PathSet) Invalidates(changev []zodb.Oid) (masters SetObject) {
	masters = SetObject{}

	for _, oid := range changev {
		roots, ok := m.residencyIdx[oid]
		if !ok {
			continue // we don't know this oid
		}

		for root := range roots {
			masters.Update(m.masterIdx[root])
		}
	}

	return masters
}
