# -*- coding: utf-8 -*-
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""test wcfs filesystem from outside as python client process"""

from __future__ import print_function

from wendelin.lib.testing import getTestDB
from wendelin.lib.zodb import dbclose
from wendelin.lib.mem import memcpy
from wendelin.bigfile.file_zodb import ZBigFile
from wendelin.bigfile.tests.test_filezodb import blksize
from wendelin import wcfs

import transaction
from persistent import Persistent
from persistent.timestamp import TimeStamp

import os, os.path, subprocess
from errno import EINVAL
from golang import func, defer
from zodbtools.util import ashex as h, fromhex
from pytest import raises

# setup:
# - create test database, compute zurl and mountpoint for wcfs
# - at every test: make sure wcfs is not running before & after the test.

testdb = None
testzurl = None     # URL of testdb
testmntpt = None    # wcfs is mounted here
def setup_module():
    global testdb, testzurl, testmntpt
    testdb = getTestDB()
    testdb.setup()

    zstor = testdb.getZODBStorage()
    testzurl = wcfs.zstor_2zurl(zstor)
    zstor.close()
    testmntpt = wcfs._mntpt_4zurl(testzurl)
    os.rmdir(testmntpt)

def teardown_module():
    testdb.teardown()

# make sure we start every test without wcfs server running.
def setup_function(f):
    assert not os.path.exists(testmntpt)

# make sure we unmount wcfs after every test.
def teardown_function(f):
    mounted = not subprocess.call(["mountpoint", "-q", testmntpt])
    if mounted:
        subprocess.check_call(["fusermount", "-u", testmntpt])
    if os.path.exists(testmntpt):
        os.rmdir(testmntpt)


# ---- tests ----

# check that zurl does not change from one open to another storage open.
def test_zurlstable():
    for i in range(10):
        zstor = testdb.getZODBStorage()
        zurl  = wcfs.zstor_2zurl(zstor)
        zstor.close()
        assert zurl == testzurl

# test that join works.
@func
def test_join():
    zurl = testzurl
    with raises(RuntimeError, match="wcfs: join .*: server not started"):
        wcfs.join(zurl, autostart=False)

    wc = wcfs._start(zurl)
    defer(wc.close)
    assert wc.mountpoint == testmntpt
    assert readfile(wc.mountpoint + "/.wcfs/zurl") == zurl
    assert os.path.isdir(wc.mountpoint + "/head")
    assert os.path.isdir(wc.mountpoint + "/head/bigfile")

    wc2 = wcfs.join(zurl, autostart=False)
    defer(wc2.close)
    assert wc2.mountpoint == wc.mountpoint

# test that join(autostart=y) works.
@func
def test_join_autostart():
    zurl = testzurl
    with raises(RuntimeError, match="wcfs: join .*: server not started"):
        wcfs.join(zurl, autostart=False)

    wc = wcfs.join(zurl, autostart=True)
    defer(wc.close)
    assert wc.mountpoint == testmntpt
    assert readfile(wc.mountpoint + "/.wcfs/zurl") == zurl
    assert os.path.isdir(wc.mountpoint + "/head")
    assert os.path.isdir(wc.mountpoint + "/head/bigfile")


# tDB is database/wcfs testing environment.
class tDB:
    def __init__(t):
        t.root  = testdb.dbopen()
        t.wc    = wcfs.join(testzurl, autostart=True)

        # commited: head + head history
        t.head   = None
        t._headv = []

        # fh(.wcfs/zhead) + history of zhead read from there
        t._wc_zheadfh = open(t.wc.mountpoint + "/.wcfs/zhead")
        t._wc_zheadv  = []

    def close(t):
        t._wc_zheadfh.close()
        t.wc.close()
        dbclose(t.root)

    # commit commits transaction and remembers/returns committed transaction ID.
    def commit(t):
        # NOTE there is no clean way to retrieve tid of just committed transaction
        #      we are using last._p_serial as workaround.
        t.root['_last'] = last = Persistent()
        last._p_changed = 1
        transaction.commit()
        head = last._p_serial
        #print('commit -> %s' % h(head))
        t.head = head
        t._headv.append(head)
        return head

    # wcsync makes sure wcfs synchronized to latest committed transaction.
    def wcsync(t):
        while len(t._wc_zheadv) < len(t._headv):
            l = t._wc_zheadfh.readline()
            #print '> zhead read: %r' % l
            l = l.rstrip('\n')
            wchead = fromhex(l)
            i = len(t._wc_zheadv)
            if wchead != t._headv[i]:
                raise RuntimeError("wcsync #%d: wczhead (%s) != zhead (%s)" % (i, wchead, t._headv[i]))
            t._wc_zheadv.append(wchead)

        # head/at = last txn of whole db
        assert t.read("head/at") == h(t.head)


    # path returns path for object on wcfs.
    # - str: wcfs root + obj;
    # - Persistent: wcfs root + (head|@<at>)/bigfile/obj
    def path(t, obj, at=None):
        if isinstance(obj, Persistent):
            head = "head/" if at is None else ("@%s/" % h(at))
            obj  = "%s/bigfile/%s" % (head, h(obj._p_oid))
            at   = None

        assert isinstance(obj, str)
        assert at is None  # must not be used with str

        return os.path.join(t.wc.mountpoint, obj)

    # read reads file corresponding to obj on wcfs.
    def read(t, obj, at=None):
        path = t.path(obj, at=at)
        return readfile(path)

    # stat stats file corresponding to obj on wcfs.
    def stat(t, obj, at=None):
        path = t.path(obj, at=at)
        return os.stat(path)

    # open opens file corresponding to obj on wcfs.
    def open(t, obj, at=None):
        path = t.path(obj, at=at)
        return open(path)

    # readblk reads ZBigFile[blk] from wcfs.
    @func
    def readblk(t, zf, blk, at=None):
        assert isinstance(zf, ZBigFile)
        f = t.open(zf, at=at)   # XXX binary, !buffered
        defer(f.close)

        blksize = zf.blksize
        f.seek(blk*blksize)
        n = blksize
        data = b''
        while n > 0:
            chunk = f.read(n)
            assert len(chunk) > 0
            data += chunk
            n    -= len(chunk)
        return data


    # assertData asserts that wcfs file corresponding to zf has data blocks as specified in expectv.
    #
    # Expected blocks may be given with size < zf.blksize. In such case they
    # are implicitly appended with trailing zeros.
    def assertData(t, zf, expectv, at=None):
        assert isinstance(zf, ZBigFile)
        data = t.read(zf, at=at)
        assert len(data) == len(expectv)*zf.blksize
        for i, blk in enumerate(expectv):
            assert len(blk) <= zf.blksize
            blk += b'\0'*(zf.blksize - len(blk))    # trailing zeros
            assert data[i*zf.blksize:(i+1)*zf.blksize] == blk



# XXX text ...
# XXX parametrize zblk0, zblk1  XXX or just rely on tox?
@func
def test_wcfs():
    t = tDB()
    defer(t.close)

    t.root['!file'] = nonfile = Persistent()
    t.root['zfile'] = f = ZBigFile(blksize)

    tid1 = t.commit()
    tid2 = t.commit()
    assert tidtime(tid2) > tidtime(tid1)
    t.wcsync()

    # lookup to non-BigFile - must be rejected
    with raises(OSError) as exc:
        t.stat(nonfile)
    assert exc.value.errno == EINVAL

    # file initially empty
    _ = t.stat(f)
    assert _.st_size  == 0
    assert _.st_mtime == tidtime(tid1)

    # commit data to f and make sure we can see it on wcfs
    # use !wcfs mode so that we prepare data independently of wcfs code paths.
    hole = 10
    fh  = f.fileh_open(_use_wcfs=False)
    vma = fh.mmap(hole, 1)    # 1 page at offset=10
    s   = b"hello world"
    memcpy(vma, s)

    t.commit()
    t.wcsync()    # sync wcfs to ZODB

    fsize = (hole + 1)*blksize  # size is always mutiple of blksize.
    _ = t.stat(f)
    assert _.st_size  == fsize
    assert _.st_mtime == tidtime(t.head)

    t.assertData(f, [b'']*hole + [s])


    for i in range(hole):
        assert t.readblk(f, i) == b'\0'*blksize
    assert t.readblk(f, hole)  == s + b'\0'*(blksize - len(s))

    data = t.read(f)
    assert len(data) == fsize
    for i in range(hole):
        assert data[i*blksize:(i+1)*blksize] == b'\0'*blksize
    tail = data[hole*blksize:]
    assert tail[:len(s)] == s
    assert tail[len(s):] == b'\0'*(blksize - len(s))


    # commit data again and make sure we can see both latest and snapshotted states.
    tcommit1 = t.head

    fh  = f.fileh_open(_use_wcfs=False)
    vma1 = fh.mmap(hole, 1)
    vma2 = fh.mmap(hole+1, 1)
    s1  = b"hello 123"
    s2  = b"alpha"
    memcpy(vma1,s1)
    memcpy(vma2,s2)

    t.commit()
    t.wcsync()

    fsize1 = fsize
    fsize  = fsize1 + blksize   # we added one more block

    _ = t.stat(f)
    assert _.st_size  == fsize
    assert _.st_mtime == tidtime(t.head)

    data = t.read(f)
    assert len(data) == fsize
    for i in range(hole):
        assert data[i*blksize:(i+1)*blksize] == b'\0'*blksize
    tail1 = data[hole*blksize:(hole+1)*blksize]
    assert tail1[:len(s1)] == s1
    assert tail1[len(s1):len(s)] == "ld"
    assert tail1[len(s):] == b'\0'*(blksize - len(s))
    tail2 = data[(hole+1)*blksize:]
    assert tail2[:len(s2)] == s2
    assert tail2[len(s2):] == b'\0'*(blksize - len(s2))


    # f @tcommit1
    st = t.stat(f, at=tcommit1)
    assert st.st_size  == fsize1
    #assert st.st_mtime == tidtime(tcommit1)     FIXME mtime for @revX -> = revX ?
    #assert readfile(fpath + "/at") == h(tcommit1)   XXX do we need it?

    data = t.read(f, at=tcommit1)
    assert len(data) == fsize1
    for i in range(hole):
        assert data[i*blksize:(i+1)*blksize] == b'\0'*blksize
    tail = data[hole*blksize:]
    assert tail[:len(s)] == s
    assert tail[len(s):] == b'\0'*(blksize - len(s))


# TODO pagecache state after loading (via mincore)
# TODO pagecahce state after .size change

def test_wcfs_invproto():
    # XXX
    pass


# ---- misc ---

# readfile reads file @ path.
def readfile(path):
    with open(path) as f:
        return f.read()

# tidtime converts tid to transaction commit time.
def tidtime(tid):
    t = TimeStamp(tid).timeTime()

    # ZODB/py vs ZODB/go time resolution is not better than 1µs
    # see e.g. https://lab.nexedi.com/kirr/neo/commit/9112f21e
    #
    # NOTE pytest.approx supports only ==, not e.g. <, so we use plain round.
    return round(t, 6)
